
// --- Definitions

const BROWSER_ACTION_IMAGES = 0;
const BROWSER_ACTION_PROXY = 1;

const MENU_ID_TAKEOVER_LISTS = 'takeover-lists';

const DATA_IMAGE_TYPES = {
  'image': true, 'imageset': true
};
const DATA_IMAGE_PROXY = 'data:image/png;base64,'
                         + 'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAABHNCSVQICAgIfAhkiAAAAAlwSFlz'
                         + 'AAAA3QAAAN0BcFOiBwAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAAANSURB'
                         + 'VAiZY2BgYGAAAAAFAAGHoU7UAAAAAElFTkSuQmCC';

const DATA_CONTENT_TYPES = {
  'main_frame': true, 'sub_frame': true, 'script': true, 'xmlhttprequest': true
};
const DATA_CONTENT_MAX = 2048*1024; // 2 MB
const GLOBAL_SUBST_MAX = 64*1024; // 64kB

const TAB_CREATION_DELAY = 1.0*1000; // 1 second


// --- Variables

var prefLogEveryRequest = false;
var prefBlockImages = DEFAULT_BLOCK_IMAGES;
var prefLogInterventions = DEFAULT_LOG_INTERVENTIONS;

var listLocations = DEFAULT_LIST_LOCATIONS;
var listContents = DEFAULT_LIST_CONTENTS;
var listImages = DEFAULT_LIST_IMAGES;
var listNavigations = DEFAULT_LIST_NAVIGATIONS;
var listHeaders = DEFAULT_LIST_HEADERS;

var tabsBlockingImages = {};
var requestContexts = {};

var tabContexts = {};
var lastManualPopupEvent = { type: 'start', timeStamp: Date.now() };

var listReferences = {
  [PREF_LIST_LOCATIONS]:   listLocations,
  [PREF_LIST_CONTENTS]:    listContents,
  [PREF_LIST_IMAGES]:      listImages,
  [PREF_LIST_NAVIGATIONS]: listNavigations,
  [PREF_LIST_HEADERS]:     listHeaders
}


// --- Queries

async function queryProxyConnection() {
  try {
    let proxySettings = await browser.proxy.settings.get({ });
    if (proxySettings && proxySettings.value.proxyType == 'manual') {
      return true;
    }
  }    
  catch (e) {}

  return false;
}

async function getCurrentTabId() {
  try {
    let tabs = await browser.tabs.query({ currentWindow: true, active: true });
    if (tabs && tabs.length > 0) {
      return tabs[0].id;
    }
  }
  catch (e) {}

  return browser.tabs.TAB_ID_NONE;
}

function queryBrowserAction() {
  if ('windows' in browser) {
    return BROWSER_ACTION_PROXY;
  }

  return BROWSER_ACTION_IMAGES;
}

function queryBlockImages(tabId) {
  if (tabId != browser.tabs.TAB_ID_NONE) {
    if (tabId in tabsBlockingImages) {
      return tabsBlockingImages[tabId];
    }
  }

  return prefBlockImages;
}

function findDescriptorByURL(url) {
  for (let i = 0; i < listDescriptors.length; ++i) {
    if (url.slice(-1 - listDescriptors[i].file.length) == '/' + listDescriptors[i].file) {
      return listDescriptors[i];
    }
  }

  return null;
}

// --- User Interface

async function pageAlert(message, tabId = browser.tabs.TAB_ID_NONE) {
  if (tabId == browser.tabs.TAB_ID_NONE) {
    tabId = await getCurrentTabId();
  }

  browser.scripting.executeScript({ target: { tabId: tabId },
    args: [ message ],
    func: (msg) => { alert(msg); }
  });
}

function updateBrowserImageAction(tabId) {
  if (tabId != browser.tabs.TAB_ID_NONE) {
    let blockImages = queryBlockImages(tabId);
    browser.browserAction.setTitle(
      { title: blockImages ? 'Load Images' : 'Block Images' });

    if ('setIcon' in browser.browserAction) {
      browser.browserAction.setIcon({
        path: blockImages ? '/icons/loadimages.svg' : '/icons/blockimages.svg'
      });
    }
  }
}

async function updateBrowserProxyAction() {
  let proxyConn = await queryProxyConnection();

  browser.browserAction.setTitle(
    { title: proxyConn ? 'Connect Directly' : 'Connect via Proxy' });

  if ('setIcon' in browser.browserAction) {
    browser.browserAction.setIcon({
      path: proxyConn ? '/icons/directconn.svg' : '/icons/proxyconn.svg'
    });
  }
}

async function updateBrowserAction(tabId) {
  switch (queryBrowserAction()) {
    case BROWSER_ACTION_PROXY:
      updateBrowserProxyAction();
      break;
    default:
      updateBrowserImageAction(tabId);
      break;
  }
}

function updatePageAction(tabId, enable) {
  if (tabId != browser.tabs.TAB_ID_NONE) {
    if (enable) {
      browser.pageAction.show(tabId);
    }
    else {
      browser.pageAction.hide(tabId);
    }
  }
}


// --- Infrastructure

function extendedReplacer(...args) {
  let replacement = args.shift();
  let groups = typeof(args[args.length - 1]) == 'object' ? args.pop() : { };

  let match = args[0];
  let captures = args.slice(1, args.length - 2);
  let offset = args[args.length - 2];
  let string = args[args.length - 1];

  let tokens = Array.from(replacement.matchAll(/\$(\$|&|`|'|\+?[0-9][0-9]?|<[A-Za-z_$][[A-Za-z0-9_$]*>)/g));
  if (!tokens.length) {
    return replacement;
  }

 function evaluate(pattern) {
    switch (pattern) {
      case '$':
        return '$';
      case '&':
        return match;
      case '`':
        return string.substring(0, offset);
      case "'":
        return string.substring(offset + match.length);
      default: break;
    }

    if (pattern.charAt(0) == '<') {
      let name = pattern.substring(1, pattern.length - 1);
      if (name in groups) {
        return groups[name];
      }

      return '';
    }

    let decode = false;
    if (pattern.charAt(0) == '+') {
      pattern = pattern.substring(1);
      decode = true;
    }

    let residue = '';
    let index = +pattern;
    if (index > captures.length && pattern.length > 1) {
      index = pattern.charCodeAt(0) - 48;
      residue = pattern.charAt(1);
    }

    if (index < 1 || index > captures.length) {
      return '$' + pattern;
    }

    let r = captures[index - 1];
    if (decode) {
      r = decodeURIComponent(r);
    }

    return r + residue;
  }

  let result = '';
  let i = 0, n = 0;

  tokens.forEach(t => {
    if (t.index > i) {
      result += replacement.substring(i, t.index);
      i = t.index;
    }

    result += evaluate(t[1]);
    i += t[0].length;
  });

  if (i < replacement.length) {
    result += replacement.substring(i);
  }

  return result;
}

async function getPageTextContent(tabId) {
  if (tabId == browser.tabs.TAB_ID_NONE) {
    tabId = await getCurrentTabId();
  }

  try {
    let result = await browser.scripting.executeScript({ target: { tabId: tabId },
      func: () => { return document.body.textContent; }
    });

    return result[0].result;
  }
  catch (e) {}

  return '';
}

function deserializeList(list, descriptor, strings) {
  list.length = 0;

  for (let i = 0; i < strings.length; ++i) {
    let cols = strings[i].split('\t');

    let item = {};
    for (let j = 0; j < descriptor.columns.length; ++j) {
      let value = null;

      if (j < cols.length) {
        if (descriptor.modifiers[j] == '$') {
          value = cols[j];
        }
        else if (cols[j] != '') {
          try {
            value = new RegExp(cols[j], descriptor.modifiers[j]);
          }
          catch (error) {
            console.error(DBTAG + 'Failed to process regex "' + cols[j] + '" in column "'
                          + descriptor.columns[j] + '" due to ' + error);
          }
        }
      }

      item[descriptor.columns[j]] = value;
    }

    list.push(item);
  }
}

function serializeList(list, descriptor) {
  let strings = [];

  for (let i = 0; i < list.length; ++i) {
    let cols = [];
    for (let j = 0; j < descriptor.columns.length; ++j) {
      let value = list[i][descriptor.columns[j]];
      let text = '';

      if (value != null) {
        if (typeof(value) == 'string') {
          text = value;
        }
        else if (value.constructor.name == 'RegExp') {
          text = value.source;
        }
      }

      cols.push(text);
    }

    strings.push(cols.join('\t'));
  }

  return strings;
}

async function loadList(list, descriptor) {
  let loadObj = { [descriptor.settings]: [] };
  try {
    let obj = await browser.storage.local.get(loadObj);
    if (Array.isArray(obj[descriptor.settings])) {
      deserializeList(list, descriptor, obj[descriptor.settings]);
      console.info(DBTAG + 'Loaded "' + descriptor.title + '" with ' + list.length
                   + ' entries via key "' + descriptor.settings + '"');
    }
    else {
      console.error(DBTAG + 'Failed to load "' + descriptor.title + '" via key "' + descriptor.settings
                    + '", data not stored as an array');
    }
  }
  catch (error) {
    console.error(DBTAG + 'Failed to load "' + descriptor.title + '" via key "' + descriptor.settings
                  + '" due to: ' + error);
  }
}

async function saveList(list, descriptor) {
  let saveObj = { [descriptor.settings]: serializeList(list, descriptor) };
  try {
    await browser.storage.local.set(saveObj);
    console.info(DBTAG + 'Saved ' + saveObj[descriptor.settings].length + ' lines of "' + descriptor.title
                 + '" via key "' + descriptor.settings + '"');
  }
  catch (error) {
    console.error(DBTAG + 'Failed to store ' + saveObj[descriptor.settings].length + ' lines of "' + descriptor.title
                  + '" via key "' + descriptor.settings + '" due to: ' + error);
  }
}

function takeoverList(url, content, tabId) {
  let descriptor = findDescriptorByURL(url);
  if (descriptor == null) {
    console.warn(DBTAG + 'Unidentified list source document "' + url.substr(url.lastIndexOf('/') + 1) + '"');

    let files = [];
    for (descriptor of listDescriptors) {
      files.push(descriptor.file);
    }

    pageAlert('Invalid DualBlock list source!'
              + '\n\n'
              + 'Only files named "' + files.join('", "') + '" are supported.',
              tabId);
    return;
  }

  let lines = content.split('\n');
  let strings = [];

  for (let i = 0; i < lines.length; ++i) {
    let line = lines[i].trim();
    if (line.length > 0 && line.charAt(0) != '#') {
      strings.push(line);
    }
  }

  let list = listReferences[descriptor.settings]
  if (!list) {
    console.error(DBTAG + 'List with key "' + descriptor.settings + '" is not accessible');
    return;
  }

  deserializeList(list, descriptor, strings);
  saveList(list, descriptor);

  pageAlert(list.length + ' entries loaded to "' + descriptor.title + '" list.', tabId);
}

function restoreLists() {
  for (let i = 0; i < listDescriptors.length; ++i) {
    let list = listReferences[listDescriptors[i].settings]
    if (list) {
      loadList(list, listDescriptors[i]);
    }
    else {
      console.error(DBTAG + 'Missing list with key "' + listDescriptors[i].settings + '"');
    }
  }
}

function getHeaderValue(httpHeaders, headerName) {
  headerName = headerName.toLowerCase();
  for (let i = 0; i < httpHeaders.length; ++i) {
    if (httpHeaders[i].name.toLowerCase() == headerName) {
      return httpHeaders[i].value.trim();
    }
  }

  return null;
}

function processHeaders(url, httpHeaders, options) {
  let headers = options.inplace ? httpHeaders : null;
  for (let i = 0; i < listHeaders.length; ++i) {
    if (listHeaders[i].url.test(url)) {
      if (headers == null) {
        headers = [];
        for (let j = 0; j < httpHeaders.length; ++j) {
          headers.push({ name: httpHeaders[j].name, value: httpHeaders[j].value });
        }
      }

      if ((listHeaders[i].header[0] == '>' && !options.response)
          || (listHeaders[i].header[0] == '<' && options.response))
       {
        let hdr = listHeaders[i].header.substring(1);
        if (listHeaders[i].match == null && listHeaders[i].subst) {
          if (prefLogInterventions) {
            console.log(DBTAG + 'Will insert HTTP header ' + hdr  + ' with value ' + listHeaders[i].subst
                        + ' in ' + (options.response ? 'response from ' : 'request for ') + url);
          }
          headers.push({ name: hdr, value: listHeaders[i].subst });
          headers.updated = true;
        }
        else {
          hdr = hdr.toLowerCase();
          for (let j = 0; j < headers.length; ) {
            if (headers[j].name.toLowerCase() == hdr) {
              if (listHeaders[i].subst == '') {
                if (listHeaders[i].match == null || listHeaders[i].match.test(headers[j].value)) {
                  if (prefLogInterventions) {
                    console.log(DBTAG + 'Will delete HTTP header ' + headers[j].name
                                + (listHeaders[i].match == null ? ''
                                   : ' with value "' + headers[j].value + '" passing ' + listHeaders[i].match.source)
                                + ' in ' + (options.response ? 'response from ' : 'request for ') + url);
                  }
                  headers.splice(j, 1);
                  headers.updated = true;
                  continue;
                }
              }
              else if (listHeaders[i].match.test(headers[j].value)) {
                let replacement = headers[j].value.replace(listHeaders[i].match, listHeaders[i].subst);
                if (prefLogInterventions) {
                  console.log(DBTAG + 'Will replace HTTP header ' + headers[j].name
                              + ' value "' + headers[j].value + '" with "' + replacement
                              + '" in ' + (options.response ? 'response from ' : 'request for ') + url);
                }
                headers[j].value = replacement;
                headers.updated = true;
              }
            }
            ++j;
          }
        }
      }
    }
  }

  if (headers && (options.inplace || headers.updated)) {
    return headers;
  }

  return null;
}

function toggleBlockImages(tabId) {
  if (tabId != browser.tabs.TAB_ID_NONE) {
    tabsBlockingImages[tabId] = !queryBlockImages(tabId);

    updateBrowserAction(tabId);

    return tabsBlockingImages[tabId];
  }

  return prefBlockImages;
}

async function toggleProxyConnection() {
  let proxySettings = { };
  try {
    proxySettings = await browser.proxy.settings.get({ });
    proxySettings = proxySettings.value;
  }
  catch (e) {}

  if (proxySettings.proxyType == 'manual') {
    browser.proxy.settings.clear({ });
    console.info(DBTAG + 'Proxy server settings released');
  }
  else {
    proxySettings.proxyType = 'manual';
    browser.proxy.settings.set({ 'value': proxySettings });
    console.info(DBTAG + 'Proxy server set to ' + proxySettings.http);
  }
}


// --- HTTP Event Handlers

function onHandleRequest(details) {
  if (prefLogEveryRequest) {
    console.log(DBTAG + 'Attempt to load ' + details.type + ' at URL ' + details.url);
  }

  if (DATA_IMAGE_TYPES[details.type]) {
    if (queryBlockImages(details.tabId)) {
      return { redirectUrl: DATA_IMAGE_PROXY };
    }

    for (let i = 0; i < listImages.length; ++i) {
      if (listImages[i].url.test(details.url)) {
        if (listImages[i].referer == null || listImages[i].referer.test(details.originUrl)) {
          if (prefLogInterventions) {
            console.log(DBTAG + 'Will be hiding image ' + details.url + ' matching ' + listImages[i].url.source
                        + (listImages[i].referer == null ? ''
                           : ' with referer ' + details.originUrl + ' complying ' + listImages[i].referer.source));
          }

          return { redirectUrl: DATA_IMAGE_PROXY };
        }
      }
    }
  }

  for (let i = 0; i < listLocations.length; ++i) {
    if (listLocations[i].url.test(details.url)) {
      if (listLocations[i].referer == null || listLocations[i].referer.test(details.originUrl)) {
        if (listLocations[i].target) {
          let redirection = details.url.replace(listLocations[i].url,
                                                (...args) => extendedReplacer(listLocations[i].target, ...args));
          if (prefLogInterventions) {
            console.log(DBTAG + 'Will redirect ' + details.type + ' URL ' + details.url + ' to ' + redirection
                        + ' by ' + listLocations[i].url.source
                        + (listLocations[i].referer == null ? ''
                           : ' with referer ' + details.originUrl + ' passing ' + listLocations[i].referer.source));
          }

          return { redirectUrl: redirection }
        }
        else {
          if (prefLogInterventions) {
            console.log(DBTAG + 'Will block ' + details.type + ' URL ' + details.url + ' by ' + listLocations[i].url.source
                          + (listLocations[i].referer == null ? ''
                             : ' with referer ' + details.originUrl + ' passing ' + listLocations[i].referer.source));
          }

          return { cancel: true };
        }
      }
    }
  }

  if (details.type == 'main_frame') {
    for (let i = 0; i < listNavigations.length; ++i) {
      if (listNavigations[i].origin.test(details.originUrl)) {
        if (listNavigations[i].aim == null || listNavigations[i].aim.test(details.url)) {
          if (prefLogInterventions) {
            console.log(DBTAG + 'Will cancel navigation from ' + details.originUrl + ' to ' + details.url
                        + ' by ' + listNavigations[i].origin.source
                        + (listNavigations[i].aim == null
                           ? '' : ' with aim passing ' + listNavigations[i].aim.source));
          }

          return { cancel: true };
        }
      }
    }
  }

  if (DATA_CONTENT_TYPES[details.type]) {
    let matches = [];
    let substs = [];

    for (let i = 0; i < listContents.length; ++i) {
      if (listContents[i].url.test(details.url)) {
        if (prefLogInterventions) {
          console.log(DBTAG + 'Will parse ' + details.type + ' content of ' + details.url + ' for ' + listContents[i].url.source);
        }

        matches.push(listContents[i].match);
        substs.push(listContents[i].subst);
      }
    }

    if (matches.length > 0) {
      let filter = browser.webRequest.filterResponseData(details.requestId);
      if (filter) {
        let context = {
          id:      details.requestId,
          url:     details.url,
          referer: details.originUrl,
          charset: null,
          length:  NaN,
          matches: matches,
          substs:  substs,
          filter:  filter,
          payload: [],
          bytes:   0,
          level:   0
        };
        requestContexts[details.requestId] = context;

        filter.ondata = (event) => {
          context.payload.push(event.data);
          context.bytes += event.data.byteLength;
        } 
        filter.onstop = (event) => {
          onHandleCompletion(context);
        }
      }
      else {
        console.error(DBTAG + 'Response filter unavailable for request ' +  details.requestId);
      }
    }
  }
}

function onHandleRequestHeaders(details) {
  if (details.requestHeaders) {
    let headers = processHeaders(details.url, details.requestHeaders, { response: false });
    if (headers) {
      return { requestHeaders: headers };
    }
  }
}

function onHandleResponseHeaders(details) {
  if (details.responseHeaders) {
    let fixCharset = false;
    if (details.requestId in requestContexts && requestContexts[details.requestId].filter) {
      let value = getHeaderValue(details.responseHeaders, 'Content-Length');
      if (value) {
        requestContexts[details.requestId].length = parseInt(value);
      }

      value = getHeaderValue(details.responseHeaders, 'Content-Type');
      if (value) {
        let params = value.split(';');
        for (let j = 0; j < params.length; ++j) {
          value = params[j].trim();
          let index = value.indexOf('=');
          if (index > 0 && value.substr(0, index).trim().toLowerCase() == 'charset') {
            let charset = value.substr(index + 1).trim().toLowerCase();
            requestContexts[details.requestId].charset = charset;
            if (charset != 'utf-8' && requestContexts[details.requestId].length < DATA_CONTENT_MAX) {
              fixCharset = true;
            }
            break;
          }
        }
      }
    }

    let headers = details.responseHeaders;
    if (fixCharset) {
      headers = [];
      for (let j = 0; j < details.responseHeaders.length; ++j) {
        let hdr = { name: details.responseHeaders[j].name, value: details.responseHeaders[j].value };
        if (hdr.name.toLowerCase() == 'content-type') {
          if (prefLogInterventions) {
            console.log(DBTAG + 'Will switch ' + details.url + ' content charset from '
                        + requestContexts[details.requestId].charset + ' to utf-8');
          }
          hdr.value = hdr.value.replace(new RegExp('(^|\\s+|;)(charset\\s*=\\s*)([0-9a-z._-]+)($|\\s+|;)', 'i'),
                                        '$1$2utf-8$4');
        }
        headers.push(hdr);
      }
    }

    headers = processHeaders(details.url, headers, { response: true, inplace: fixCharset });
    if (headers) {
      return { responseHeaders: headers };
    }
  }
}

function onHandleResponse(details) {
  if (details.requestId in requestContexts && requestContexts[details.requestId].filter) {
    onHandleCompletion(requestContexts[details.requestId]);
  }
}

function onHandleCompletion(context) {
  context.level += 1;

  if (context.level >= 2) {
    if (context.bytes <= DATA_CONTENT_MAX) {
      let charset = context.charset || 'utf-8';
      let done = false;
      let data = '';
      for (let i = 0; i < context.matches.length; ++i) {
        if (context.matches[i] == null) {
          if (prefLogInterventions) {
            console.log(DBTAG + 'Will set ' + context.bytes + 'B of content at ' + context.url);
          }

          data = context.substs[i] == null ? '' : context.substs[i];
          done = true;
          break;
        }
      }

      if (!done) {
        let decoder = new TextDecoder(charset);
        for (let i = 0; i < context.payload.length; ++i) {
          data += decoder.decode(context.payload[i], { stream: i + 1 < context.payload.length });
        }

        for (let i = 0; i < context.matches.length; ++i) {
          if (context.matches[i].test(data)) {
            if (prefLogInterventions) {
              console.log(DBTAG + 'Will be replacing in ' + data.length + ' chars text at ' + context.url
                          + ' matching ' + context.matches[i].source);
            }

            if (context.substs[i] == null) {
              data = '';
              break;
            }

            let match = context.matches[i];
            if (data.length < GLOBAL_SUBST_MAX) {
              match = new RegExp(match.source, 'g' + match.flags)
            }

            data = data.replace(match, context.substs[i]);
          }
        }
      }

      let encoder = new TextEncoder();
      let buffer = encoder.encode(data);
      context.filter.write(buffer);
    }
    else {
      console.warn(DBTAG + 'Content of ' + context.url + ' exceeds ' + DATA_CONTENT_MAX + 'B');
      for (let i = 0; i < context.payload.length; ++i) {
        context.filter.write(context.payload[i]);
      }
    }

    delete requestContexts[context.id];
    context.filter.close();
  }
}


// --- Infrastructure Event Handlers

function onHandleProxyChange(details) {
  updateBrowserAction(browser.tabs.TAB_ID_NONE);
}

function onHandleStorageChange(changes) {
  if (PREF_BLOCK_IMAGES in changes) {
    prefBlockImages = changes[PREF_BLOCK_IMAGES].newValue;
  }

  if (PREF_LOG_INTERVENTIONS in changes) {
    prefLogInterventions = changes[PREF_LOG_INTERVENTIONS].newValue;
  }

  for (let i = 0; i < listDescriptors.length; ++i) {
    if (listDescriptors[i].settings in changes) {
      let list = listReferences[listDescriptors[i].settings]
      if (list) {
        loadList(list, listDescriptors[i]);
      }
      else {
        console.error(DBTAG + 'List "' + listDescriptors[i].settings + '" is unaccessible');
      }
    }
  }
}

function onHandleMessage(message, sender, sendResponse) {
  if (message && message.type) {
    switch (message.type) {
      case 'mouseup': {
        if (message.button == 1) {
          lastManualPopupEvent = message;
          lastManualPopupEvent.originUrl = sender.tab ? sender.tab.url : null;
        }
        break;
      }
    }
  }
}


// --- GUI Event Handlers

async function onHandleWindowFocusChanged(windowId) {
  if (windowId != browser.windows.WINDOW_ID_NONE) {
    let tabs = await browser.tabs.query({ currentWindow: true, active: true });
    if (tabs.length > 0) {
      updatePageAction(tabs[0].id, findDescriptorByURL(tabs[0].url) != null);
      updateBrowserAction(tabs[0].id);
    }
  }
}

function onHandleTabCreated(tab) {
  if (tab.id != browser.tabs.TAB_ID_NONE) {
    let context = {
      id:         tab.id,
      created:    Date.now(),
      unintended: false
    }

    if (context.created - lastManualPopupEvent.timeStamp > TAB_CREATION_DELAY) {
      context.unintended = true;
    }
    tabContexts[tab.id] = context;

    if (prefLogEveryRequest) {
      console.log(DBTAG + 'Tab #' + tab.id + ' created' + (context.unintended ? ' and marked unintended' : ''));
    }
  }
}

function onHandleTabRemoved(tabId, removeInfo) {
  delete tabContexts[tabId];

  if (prefLogEveryRequest) {
    console.log(DBTAG + 'Tab #' + tabId + ' closed');
  }
}

async function onHandleTabActivated(activeInfo) {
  let tabs = await browser.tabs.query({ currentWindow: true, active: true });
  if (tabs.length > 0) {
    updatePageAction(tabs[0].id, findDescriptorByURL(tabs[0].url) != null);
    updateBrowserAction(tabs[0].id);
  }
}

function onHandleTabUpdated(tabId, changeInfo, tab) {
  if (changeInfo.status) {
    updatePageAction(tabId, changeInfo.status == 'complete'
                            && findDescriptorByURL(tab.url) != null);
  }

  if (changeInfo.url) {
    if (!(tabId in tabContexts)) {
      tabContexts[tabId] = { id: tab.id }
    }

    tabContexts[tabId].url = changeInfo.url;
  }

  if (tabId in tabContexts && tabContexts[tabId].unintended && tabContexts[tabId].originUrl) {
    for (let i = 0; i < listNavigations.length; ++i) {
      if (listNavigations[i].origin.test(tabContexts[tabId].originUrl)) {
        if (listNavigations[i].close == null
            || (tabContexts[tabId].url && listNavigations[i].close.test(tabContexts[tabId].url))) {
          if (prefLogInterventions) {
            console.log(DBTAG + 'Will close popup tab #' + tabId + ' originating in '
                        + tabContexts[tabId].originUrl + ' by ' + listNavigations[i].origin.source
                        + (listNavigations[i].close == null ? '' : ' with aim passing ' + listNavigations[i].close.source));
          }

          browser.tabs.remove(tabId);
        }
      }
    }
  }
}

async function onHandleCreatedNavigationTarget(details) {
  if (details.tabId in tabContexts) {
    tabContexts[details.tabId].url = details.url;

    if (details.sourceTabId != browser.tabs.TAB_ID_NONE) {
      let tab = await browser.tabs.get(details.sourceTabId);
      if (tab.url) {
        tabContexts[details.tabId].originUrl = tab.url;
        if (prefLogInterventions) {
          console.log(DBTAG + 'Tab #' + details.tabId + ' was created on behalf of ' + tab.url
                      + (tabContexts[details.tabId].unintended ? ' and marked unintended' : ''));
        }
      }
    }
  }
}

async function onHandleMenu(info, tab) {
  if (info.menuItemId == MENU_ID_TAKEOVER_LISTS) {
    let text = await getPageTextContent(tab.id);
    takeoverList(tab.url, text, tab.id);
  }
}

async function onHandleBrowserAction(tab, data) {
  switch (queryBrowserAction()) {
    case BROWSER_ACTION_PROXY:
      toggleProxyConnection();
      break;
    default: {
      let tabs = await browser.tabs.query({ currentWindow: true, active: true });
      if (tabs.length > 0) {
        if (!toggleBlockImages(tabs[0].id)) {
          browser.tabs.reload(tabs[0].id, { bypassCache: true });
        }
      }
      break;
    }
  }
}

async function onHandlePageAction(tab) {
  let text = await getPageTextContent(tab.id);
  takeoverList(tab.url, text, tab.id);
}


// --- Initialization Code

async function initialize() {
  let loadObj = {
    [PREF_BLOCK_IMAGES]:      prefBlockImages,
    [PREF_LOG_INTERVENTIONS]: prefLogInterventions
  };

  let obj = await browser.storage.local.get(loadObj);
  prefBlockImages = obj[PREF_BLOCK_IMAGES];
  prefLogInterventions = obj[PREF_LOG_INTERVENTIONS];

  let tabId = await getCurrentTabId();
  updateBrowserAction(tabId);
}


// --- Main

initialize();
restoreLists();

browser.proxy.settings.onChange.addListener(onHandleProxyChange);
browser.storage.onChanged.addListener(onHandleStorageChange);
browser.runtime.onMessage.addListener(onHandleMessage);

if ('windows' in browser) {
  browser.windows.onFocusChanged.addListener(onHandleWindowFocusChanged);
}

browser.tabs.onCreated.addListener(onHandleTabCreated);
browser.tabs.onRemoved.addListener(onHandleTabRemoved);
browser.tabs.onActivated.addListener(onHandleTabActivated);
browser.tabs.onUpdated.addListener(onHandleTabUpdated);

browser.browserAction.onClicked.addListener(onHandleBrowserAction);
browser.pageAction.onClicked.addListener(onHandlePageAction);

if ('menus' in browser) {
  browser.menus.create(
    { id: MENU_ID_TAKEOVER_LISTS, title: 'Takeover DB Lists',
      icons: { 16: "/icons/regexreload.svg" }, contexts: [ 'tools_menu' ]
  });

  browser.menus.onClicked.addListener(onHandleMenu);
}

browser.webNavigation.onCreatedNavigationTarget.addListener(onHandleCreatedNavigationTarget);

browser.webRequest.onBeforeRequest.addListener(onHandleRequest,
  { urls: [ '<all_urls>' ] }, [ 'blocking' ]);
browser.webRequest.onBeforeSendHeaders.addListener(onHandleRequestHeaders,
  { urls: [ '<all_urls>' ] }, [ 'blocking', 'requestHeaders' ]);
browser.webRequest.onHeadersReceived.addListener(onHandleResponseHeaders,
  { urls: [ '<all_urls>' ] }, [ 'blocking', 'responseHeaders' ]);
browser.webRequest.onResponseStarted.addListener(onHandleResponse,
  { urls: [ '<all_urls>' ] }, [ 'responseHeaders' ]);
